#include <stdio.h>
#include <ncurses.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>

/**************************************
***************************************
**                                   **
**          Wolfenstain  2D          **
**        /+/   Ravodin & ...        **
** Games for Agat and IBM PC NetWare **
**          (c)  1996, 2019          **
**                                   **
***************************************
**************************************/

#define SM_NETMODE	(0)
#define SM_SINGLEMODE	(1)
int single_mode;			// 0 -  ; 1 -  ,    

#define NM_SERVER	(0)
#define NM_CLIENT	(1)
int network_mode;			// 0 - , ,   ; 1 - ,  

const int maxx = 64, maxy = 32;		//   

int dead_cnt = 0, dead_frend_cnt = 0;	// 

					//  
const int XViewLeft = 8, XViewRight = 8, YViewUp = 4, YViewDown = 4;

int fd_socket = -1;			// FD UDP    

u_int8_t map[maxx][maxy];		//  

#pragma pack(push)
#pragma pack(1)
typedef struct GAME_STATE {
  u_int8_t	sign;
  u_int8_t	x, y, angle;
  u_int8_t	rocket_x, rocket_y, rocket_angle;
} GAME_STATE;
#pragma pack(pop)                                                                                                   

GAME_STATE local_state, prev_local_state, remote_state;


int col[8][8];

static void initcol() {
  int i, j, c = 1,
    c_ind[8]={COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN, 
              COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE};

  start_color();
  for (i = 0; i <= 7; i++) for (j = 0; j <= 7; j++) if (i == j) {
    col[i][j] = 0;
  } else {
    if (c > COLOR_PAIRS-1) { printf("COLOR_PAIRS fault (max = %d)\n", COLOR_PAIRS); exit(12); }
    init_pair(c, c_ind[j], c_ind[i]);
    col[i][j] = c;
    c++;
  }
}


void setcolor(unsigned char c) {
  if (c & 0x08) { attron(A_BOLD); } else { attroff(A_BOLD); };
  attron( COLOR_PAIR( col[ (c >> 4) & 0x07 ] [ (c     ) & 0x07 ] ) );
}


void open_socket(char * local_port, char * remote_host, char * remote_port) {
  struct addrinfo hints, *servinfo, *p;
  int err;

  memset(&hints, 0, sizeof hints);
  hints.ai_family   = AF_INET;		// ip4v only
  hints.ai_socktype = SOCK_DGRAM;	// UDP
  hints.ai_flags    = AI_PASSIVE;	// MyIP use

  err = getaddrinfo(NULL, local_port, &hints, &servinfo);
  if (err != 0) { printf("Getaddrinfo 1: %s\n", gai_strerror(err)); exit(1); }

  //    ,    
  for (p = servinfo; p != NULL; p = p->ai_next) {
    if ((fd_socket = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) { perror("Socket"); continue; }

    // bind     ,     
    if (bind(fd_socket, p->ai_addr, p->ai_addrlen) == -1) { perror("Bind"); close(fd_socket); continue; }

    break;
  }

  if (p == NULL) { printf("Failed to bind socket 1\n"); exit(1); }

  freeaddrinfo(servinfo);

// -- //

  memset(&hints, 0, sizeof hints);
  hints.ai_family   = AF_INET;		// ip4v only
  hints.ai_socktype = SOCK_DGRAM;	// UDP
//  hints.ai_flags    = AI_PASSIVE;	// MyIP use

  err = getaddrinfo(remote_host, remote_port, &hints, &servinfo);
  if (err != 0) { printf("Getaddrinfo 2: %s\n", gai_strerror(err)); exit(1); }

  for (p = servinfo; p != NULL; p = p->ai_next) {
    if (connect(fd_socket, p->ai_addr, p->ai_addrlen) != -1) break;
    perror("Connect");
  }

  if (p == NULL) { printf("Failed to bind socket 2\n"); exit(1); }

  freeaddrinfo(servinfo);
}


int rand_range(int range) {
  double t = (double) (random() & 0xFFFFFF) / 0xFFFFFF;
  return t * range;
}

void respawn_in_new_place() {
  do {
    local_state.x = rand_range(maxx);
    local_state.y = rand_range(maxy);
    local_state.angle = rand_range(2) * 2;
  } while (map[local_state.x][local_state.y] > 0);
};


void transmit_map() {
  u_int8_t lmap[maxx];
  int rsize = sizeof(lmap);

  if (network_mode == NM_SERVER) {
    for (int y = 0; y < maxy; y++) {
      for (int x = 0; x < maxx; x++) lmap[x] = map[x][y];
      int res = write(fd_socket, lmap, rsize);
      if (res < 0) { perror("write"); exit(1); };
      if (res != rsize) { printf("Write map: size fault %d < %d\n", res, rsize); exit(1); };
      usleep(50000);
    };
    printf("Map flyed to remote...\n");
  } else {
    printf("Wait map from remote...\n");
    for (int y = 0; y < maxy; y++) {
      int res = read(fd_socket, &lmap, rsize);
      for (int x = 0; x < maxx; x++) map[x][y] = lmap[x];
      if (res < 0) { perror("read"); exit(1); };
      if (res != rsize) { printf("Read map: size fault %d < %d\n", res, rsize); exit(1); };
    };
    printf("Map received !\n");
  }
};


// ޣ  
void calc_view_window(GAME_STATE st, int *x1, int *x2, int *y1, int *y2) {
  switch (st.angle) {
    case 0: *x1 = st.x-XViewLeft;	*x2 = st.x+XViewRight;	*y1 = st.y-YViewUp;	*y2 = st.y;		break;
    case 2: *x1 = st.x;			*x2 = st.x+XViewRight;	*y1 = st.y-YViewUp;	*y2 = st.y+YViewDown;	break;
    case 4: *x1 = st.x-XViewLeft;	*x2 = st.x+XViewRight;	*y1 = st.y;		*y2 = st.y+YViewDown;	break;
    case 6: *x1 = st.x-XViewLeft;	*x2 = st.x;		*y1 = st.y-YViewUp;	*y2 = st.y+YViewDown;	break;
  };

  if (*x1 < 0)		*x1 = 0;
  if (*x2 >= maxx)	*x2 = maxx - 1;
  if (*y1 < 0)		*y1 = 0;
  if (*y2 >= maxy)	*y2 = maxy - 1;
};


//    
void draw_player(GAME_STATE st, int color) {
  setcolor(color);
  addch("^>v<" [st.angle >> 1]);
}


//    
void game_draw_screen() {
  int x, y, x1, x2, y1, y2;

  calc_view_window(local_state, &x1, &x2, &y1, &y2);

  for (y = y1; y <= y2; y++) {
    move(y, x1); 
    for (x = x1; x <= x2; x++) {
      if ((local_state.x == x) && (local_state.y == y)) { draw_player(local_state, 14); continue; };
      if ((remote_state.x == x) && (remote_state.y == y)) { draw_player(remote_state, 10); continue; };
      if ((local_state.rocket_x == x) && (local_state.rocket_y == y) && (local_state.rocket_angle < 10)) {
	setcolor(3); addch('+'); continue;
      }
      if ((remote_state.rocket_x == x) && (remote_state.rocket_y == y) && (remote_state.rocket_angle < 10)) {
	setcolor(3); addch('*'); continue;
      }

      switch ( map[x][y] ) {
	case 0: setcolor(7); addch(' '); break;
	case 1: setcolor(7); addch(ACS_CKBOARD); break;
      };
    }; // x
  }; // y
}


//   
void clearbar(int x1, int y1, int x2, int y2) {
  int x, y;
  char s[128];
  for (x = x1; x <= x2; x++) s[x-x1] = ' ';
  s[x2-x1+1] = '\0';

  for (y = y1; y <= y2; y++) mvprintw(y, x1, "%s", s);
}


//   ,    - 
void game_clear_screen() {
  int x1, x2, y1, y2;
  
  if ((prev_local_state.x != local_state.x) || (prev_local_state.y != local_state.y)) {
    if (prev_local_state.angle == local_state.angle) {
      calc_view_window(prev_local_state, &x1, &x2, &y1, &y2);

      switch (prev_local_state.angle) {
	case 0: clearbar(x1, y2, x2, y2); break;
	case 2: clearbar(x1, y1, x1, y2); break;
	case 4: clearbar(x1, y1, x2, y1); break;
	case 6: clearbar(x2, y1, x2, y2); break;
      } // switch
    } else {
      clear();
    };
  }; // if
};


//  
void draw_stat() {
  setcolor(10); mvprintw(40, 66, "You killed: %d", dead_frend_cnt);
  setcolor(12); mvprintw(43, 67, "Your died: %d", dead_cnt);
}

#define IS_FREE_CELL(xx, yy) \
  ((map[xx][yy] == 0) && ((xx != remote_state.x) || (yy != remote_state.y)))

// ޣ   
int game_step_processing() {
  int key = getch();
//  mvprintw(1, 1, "Key = %d  ", key);
  switch ( key ) {
    case ERR: usleep(100000); break;

    case KEY_F(10):	return 0;
	//  -   
    case KEY_LEFT:	local_state.angle = 6; if (IS_FREE_CELL(local_state.x-1, local_state.y)) local_state.x--; break;
    case KEY_RIGHT:	local_state.angle = 2; if (IS_FREE_CELL(local_state.x+1, local_state.y)) local_state.x++; break;
    case KEY_UP:	local_state.angle = 0; if (IS_FREE_CELL(local_state.x, local_state.y-1)) local_state.y--; break;
    case KEY_DOWN:	local_state.angle = 4; if (IS_FREE_CELL(local_state.x, local_state.y+1)) local_state.y++; break;
	//  -    (strafe)
    case 'a':	if (IS_FREE_CELL(local_state.x-1, local_state.y)) local_state.x--; break;
    case 'd':	if (IS_FREE_CELL(local_state.x+1, local_state.y)) local_state.x++; break;
    case 'w':	if (IS_FREE_CELL(local_state.x, local_state.y-1)) local_state.y--; break;
    case 's':	if (IS_FREE_CELL(local_state.x, local_state.y+1)) local_state.y++; break;
   	// 
    case 'q':	local_state.angle = (7 & (local_state.angle - 2)); break;
    case 'e':	local_state.angle = (7 & (local_state.angle + 2)); break;
	// 
    case ' ':	if (local_state.rocket_angle > 10) {
		  local_state.rocket_angle = local_state.angle;
		  local_state.rocket_x = local_state.x;
		  local_state.rocket_y = local_state.y;
	        };
		break;
  };

  //  
  if (
       (remote_state.rocket_x == local_state.x) &&
       (remote_state.rocket_y == local_state.y) &&
       (remote_state.rocket_angle < 10)
     ) {
    setcolor(14);
    mvprintw(local_state.y, local_state.x-1, "RIP");
    sleep(1);
    dead_cnt++;
    draw_stat();
    respawn_in_new_place();
  };

  //  
  if ((local_state.rocket_x == remote_state.x) && (local_state.rocket_y == remote_state.y)) {
    setcolor(14);
    mvprintw(local_state.y, local_state.x-1, "8-E");
    dead_frend_cnt++;
    draw_stat();
  };

  //   
  if (local_state.rocket_angle < 10) {
    switch (local_state.rocket_angle) {
      case 0: local_state.rocket_y--; break;
      case 2: local_state.rocket_x++; break;
      case 4: local_state.rocket_y++; break;
      case 6: local_state.rocket_x--; break;
    };
    if (map[local_state.rocket_x][local_state.rocket_y] > 0) local_state.rocket_angle = 128;
  };


  return 1;
};


void game_step_send() {
  int ret;
  int rsize = sizeof(GAME_STATE);

  local_state.sign = 31;
  ret = write(fd_socket, &local_state, rsize);
  if (ret < 0) { perror("Write"); exit(1); };
  if (ret != rsize) { printf("Write: size fault %d < %d\n", ret, rsize); exit(1); };
}

void game_step_recv() {
  int ret;
  int rsize = sizeof(GAME_STATE);
  GAME_STATE tmp;

  //   ,    -   
  while (1) {
    ret = recv(fd_socket, &tmp, rsize, MSG_DONTWAIT);
    if (ret < 0) {
      if (errno == EAGAIN) return; else { perror("Recv"); exit(1); };
    }
    if (ret == 0) { printf("Recv: remote close socket\n"); exit(0); };
    if (ret != rsize) { printf("Recv: size fault %d < %d\n", ret, rsize); exit(1); };
    if (tmp.sign == 31) remote_state = tmp;
  };
}


//    ̣ 
void game_step_exchange() {
  if (network_mode == NM_SERVER) {
    game_step_send();
    game_step_recv();
  } else {
    game_step_recv();
    game_step_send();
  };
};


void close_curses() {
  curs_set(1);
  endwin();
  if (fd_socket >= 0) close(fd_socket);
}


int main(int argc, char *argv[]) {
  switch (argc) {
    case 2:
      single_mode = SM_SINGLEMODE;
      break;

    case 4:
      single_mode = SM_NETMODE;
      network_mode = NM_CLIENT;
      break;

    case 5:
      single_mode = SM_NETMODE;
      network_mode = NM_SERVER;
      break;

    default:
      printf("Local port, remote addr and remote port required\n");
      exit(1);
  }

  char map_filename[128] = "\0";

  if (single_mode == SM_SINGLEMODE) {
    strncpy(map_filename, argv[1], sizeof(map_filename));
  } else {
    if (network_mode == NM_SERVER) {
      strncpy(map_filename, argv[4], sizeof(map_filename));
    }
    open_socket(argv[1], argv[2], argv[3]);
  }

  if (map_filename[0]) {
    int fd = open(map_filename, O_RDONLY);
    if (fd < 0) { perror("Load map"); exit(1); }
    if ( read(fd, map, sizeof(map)) < 0 ) { perror("Load map"); exit(1); }
    close(fd);
  }

  if (single_mode == SM_NETMODE) transmit_map();

  respawn_in_new_place();

  local_state.rocket_angle = 128;
  remote_state.x = 200;			//     ,   ,
  remote_state.y = 200;			//   ,   -   

  initscr(); atexit(&close_curses);	//  
  noecho();				// 
  clear();				//    
  keypad(stdscr, TRUE);			//    ncurses
  raw();				// ^C -                                                          
  nodelay(stdscr, 1);			// getch   ERR,    
  curs_set(0);				//  
  // cbreak();				//  . Ctrl-c 
  //getmaxyx(stdscr, maxy, maxx);	//   
  initcol();				//   col -  

  do {
    game_draw_screen();
    prev_local_state = local_state;
    if (! game_step_processing()) exit(0);
    if (single_mode == SM_NETMODE) game_step_exchange();
    game_clear_screen();
  } while (1);
}
